/* webserver.c - Serving files through the libupnp built in web server
 *
 * Copyright (C) 2005  Oskar Liljeblad
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <config.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include "xalloc.h"
#include "intutil.h"
#include "gmediaserver.h"

typedef struct {
    int fd;
    char *fullpath;
    Entry *entry;
} WebServerFile;

static Entry *
get_file_entry(const char *filename)
{
    int32_t id;

    if (filename[0] != '/')
	return NULL;
    filename = strchr(filename+1, '/');
    if (filename == NULL)
	return NULL;
    filename++;
    if (filename[0] == '\0')
	return NULL;
    if (!parse_int32(filename, &id))
	return NULL;

    return get_entry_by_id(id);
}

static int
webserver_get_info(const char *filename, struct File_Info *info)
{
    Entry *entry;
    char *fullpath;
    struct stat sb;

    say(4, "Request file information for %s\n", filename);

    entry = get_file_entry(filename);
    if (entry == NULL)
	return -1;

    fullpath = get_entry_path(entry);
    if (stat(fullpath, &sb) < 0) {
        warn("%s: cannot stat: %s\n", fullpath, errstr);
	free(fullpath);
	return -1;
    }

    if (access(fullpath, R_OK) < 0) {
	if (errno != EACCES) {
	    warn("%s: cannot check access: %s\n", fullpath, errstr);
	    free(fullpath);
	    return -1;
	}
	info->is_readable = 0;
    } else {
	info->is_readable = 1;
    }
    info->file_length = sb.st_size;
    info->last_modified = sb.st_mtime;
    info->is_directory = S_ISDIR(sb.st_mode);
    if (info->is_directory) {
	/* XXX: not really sure what content type to set for directories */
	info->content_type = ixmlCloneDOMString("");
    } else {
	/* XXX: safe guess always?! */
	info->content_type = ixmlCloneDOMString("audio/mpeg");
    }

    say(4, "Returned info for file %s successfully\n", fullpath);

    return 0;
}

static UpnpWebFileHandle
webserver_open(const char *filename, enum UpnpOpenFileMode mode)
{
    Entry *entry;
    int fd;
    char *fullpath;
    WebServerFile *file;

    if (mode != UPNP_READ) {
	warn("%s: ignoring request to open file for writing\n", filename);
	return NULL;
    }

    entry = get_file_entry(filename);
    if (entry == NULL)
	return NULL;

    fullpath = get_entry_path(entry);

    fd = open(fullpath, O_RDONLY);
    if (fd < 0) {
	warn("%s: cannot open for reading: %s\n", fullpath, errstr);
	free(fullpath);
	return NULL;
    }

    file = xmalloc(sizeof(WebServerFile));
    file->entry = entry;
    file->fd = fd;
    file->fullpath = fullpath;

    return (UpnpWebFileHandle) file;
}

static int
webserver_read(UpnpWebFileHandle fh, char *buf, size_t buflen)
{
    WebServerFile *file = (WebServerFile *) fh;
    ssize_t len;

    say(4, "Attempting to read %d bytes from %s\n", buflen, file->fullpath);

    len = read(file->fd, buf, buflen);
    if (len < 0)
	warn("%s: cannot read: %s\n", file->fullpath, errstr);

    return len;
}

static int
webserver_write(UpnpWebFileHandle fh, char *buf, size_t buflen)
{
    WebServerFile *file = (WebServerFile *) fh;

    warn("%s: ignoring request to write to file\n", file->fullpath);

    return -1;
}

static int
webserver_seek(UpnpWebFileHandle fh, long offset, int origin)
{
    WebServerFile *file = (WebServerFile *) fh;

    say(4, "Attempting to seek to %ld (origin %d) in %s\n", offset, origin, file->fullpath);

    if (lseek(file->fd, offset, origin) == -1) {
	warn("%s: cannot seek: %s\n", file->fullpath, errstr);
	return -1;
    }

    return 0;
}

static int
webserver_close(UpnpWebFileHandle fh)
{
    WebServerFile *file = (WebServerFile *) fh;

    close(file->fd); /* Ignore errors since only reading */
    free(file->fullpath);
    free(file);

    return 0;
}

struct UpnpVirtualDirCallbacks virtual_dir_callbacks =
{
    webserver_get_info,
    webserver_open,
    webserver_read,
    webserver_write,
    webserver_seek,
    webserver_close
};
